#/usr/bin/perl
#
#  whisker 2.0, copyright rain forest puppy / rfp(at)wiretrip.net
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#

### BEGIN CONFIG SECTION ###############################################
#
# If you have installed whisker, then indicate the appropriate install dir

if($^O=~/Win32/){
	$WHISKER_DIR='C:\\program files\\whisker\\'; 
} else {
	$WHISKER_DIR='/usr/local/lib/whisker/'; 
}


# Set to 0 if you don't want to see the "Whisker not officially
# installed.." message

$WHISKER_INSTALL_WARNING = 1;

### END CONFIG SECTION ##################################################

select(STDERR); $|++; select(STDOUT); $|++;

my $LW=0; ### This code checks to see if libwhisker is available
eval { use LW; };
if($@){	$LW=0;
    if(-e $WHISKER_DIR.'LW.pm'){
	eval{ require $WHISKER_DIR.'LW.pm'; };
	if(!$@){$LW++;}}
} else { $LW++; }
if(!$LW){
    print "ERROR:\tLibwhisker not found/installed.\n";
    print "\tDownload libwhisker from www.wiretrip.net/rfp/\n\n";
    exit;
}

my @V=split(/\./,$LW::VERSION);
if($V[0]!=1 && $V[1]<4){
    print "ERROR:\tYou're using an outdated libwhisker version\n";
    print "\tPlease download a newer version from www.wiretrip.net/rfp/\n\n";
    exit;
}

#########################################################################

$G_VERSION='2.0';

# global variables and schtuff
@G_FOUND=();	%G_HOOKS=();	%G_TRACK=();	%G_ARRAY=();
%G_REQ=();	%G_IDS=();	%G_VARBS=();	%G_RECORD=();
%G_RESP=();

#########################################################################

_init();

# find where we stashed all the plugins and tests
if(-e $WHISKER_DIR && -d $WHISKER_DIR){
	opendir(DIR,$WHISKER_DIR)||die("Can't open whisker install dir $WHISKER_DIR");
	$G_VARBS{'WHISKER_DIR'}=$WHISKER_DIR;
} else {
	print STDERR "[ Whisker not officially installed; reading from current directory ]\n"
		if($WHISKER_INSTALL_WARNING);
	opendir(DIR,'.')||die("Can't open current directory for reading");
	$G_VARBS{'WHISKER_DIR'}='./';
}

while(my $file=readdir(DIR)){
	require("$G_VARBS{'WHISKER_DIR'}$file") if($file=~m#\.plugin$#);
}


#########################################################################

my %G_O;
if(LW::utils_getopts($G_VARBS{'options'},\%G_O)){
	die("Error parsing options"); }

if(scalar keys %G_O == 0){
	print STDOUT "\n--/ whisker $G_VERSION / www.wiretrip.net /";
	print STDOUT '-'x38,"\n",$G_VARBS{'usage'};
	print STDOUT " -T <arg>  Modify the scan via the tweaks specified:\n";
	print STDOUT $G_VARBS{'tweaks_usage'},"\n\n";
	exit; }

foreach (@{$G_HOOKS{'OPTIONS'}}){
	&$_(\%G_O); }

if(!defined $G_O{h}){
	print STDOUT "ERROR: you didn't specify a host to scan.\n";
	exit; }

$G_VARBS{'target'}=$G_O{h};
$G_VARBS{'target'}='http://'.$G_VARBS{'target'} 
	if($G_VARBS{'target'}!~m#^[a-zA-Z]+://#);

if($G_VARBS{'targate'}=~m#^https://#i && $LW::LW_HAS_SSL==0){
	print STDOUT <<EOT;
Error: you requested an HTTPS site, but SSL support is not available.  You
need to install the Net::SSLeay or Net::SSL Perl module first, which
libwhisker will then automatically use for SSL scanning.
EOT
	exit;
}

if(defined $G_O{I}){ # turn on anti-ids stuff
	$G_REQ{'User-Agent'}='Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; T312461)';
	$G_VARBS{'ids_modes'}=$G_O{I}; }

if(defined $G_O{t}){
	$G_O{t}=~tr/0-9//cd;
	$G_REQ{whisker}->{timeout}=$G_O{t}||10; }

if(defined $G_O{a}){
	my @p=split(':',$G_O{a},2);
	LW::auth_set_header('basic',\%G_REQ,$p[0],$p[1]); }

if(defined $G_O{A}){
	my @p=split(':',$G_O{A},2);
	LW::auth_set_header('proxy-basic',\%G_REQ,$p[0],$p[1]); }

wtarget($G_VARBS{'target'});

if(defined $G_O{l}){
	open(G_OUT,">$G_O{l}")||die("Can't open $G_O{l} for output");
	select(G_OUT); $|++;
	$G_VARBS{'OUT'}=select(STDOUT);
} else {
	$G_VARBS{'OUT'}=select(STDOUT); $|++;
}

if(defined $G_O{r}){
	my @tests=split(/:/,$G_O{t});
	foreach (@tests){
	 	$G_VARBS{'RUN_TESTS'}->{$_}=1; }
	$G_VARBS{'default_test_exec'}=0;
}

if(defined $G_O{G}){
	$G_VARBS{'alldirs'}++;
	$G_VARBS{'allservers'}++;
}

if(defined $G_O{P}){
	my @p=split(/:/,$G_O{P});
	$p[0]=~tr/-a-zA-Z0-9.//cd;
	$p[1]=~tr/0-9.//cd;
	$G_REQ{whisker}->{'proxy_host'}=$p[0];
	$G_REQ{whisker}->{'proxy_port'}=$p[1];
}

if(defined $G_O{T}){ # handle the 'tweaks'
  $G_VARBS{'save_cookies'}++			if($G_O{T}=~/C/||$G_O{T}=~/A/);
  $G_VARBS{'track_cookies'}++ 			if($G_O{T}=~/c/);
  push(@{$G_ARRAY{'php_exts'}},'.phtml')	if($G_O{T}=~/p/);
  push(@{$G_HOOKS{EXIT}},\&_dump_track)		if($G_O{T}=~/T/);
  $G_VARBS{'rescan_lowercase'}++ 		if($G_O{T}=~/l/);
  $G_VARBS{'skip_disclaimer'}++			if($G_O{T}=~/d/);
  $G_VARBS{'default_test_exec'}=1		if($G_O{T}=~/r/);
  $G_VARBS{'DEBUG'}=1				if($G_O{T}=~/D/);
  $G_REQ{whisker}->{retry}=0			if($G_O{T}=~/R/);
  $G_REQ{whisker}->{'http_ver'}='1.0'		if($G_O{T}=~/0/);
  $G_VARBS{'save_moves'}++			if($G_O{T}=~/3/||$G_O{T}=~/A/);
  $G_VARBS{'report_auth'}++			if($G_O{T}=~/4/||$G_O{T}=~/A/);
  $G_VARBS{'report_fail'}++			if($G_O{T}=~/5/||$G_O{T}=~/A/);
  $G_VARBS{'report_ftargets'}++			if($G_O{T}=~/f/||$G_O{T}=~/A/);
  $G_VARBS{'no_interactive'}++			if($G_O{T}=~/i/);
}

if(defined $G_O{c}){ # crawl options
  $G_VARBS{'skip_spiderdirs'}++			if($G_O{c}=~/a/);
  $G_VARBS{'skip_spider'}++			if($G_O{c}=~/d/);
  $G_VARBS{'migrate_formtargdirs'}++		if($G_O{c}=~/c/);
  $G_VARBS{'skipaspphp_formtargdirs'}++		if($G_O{c}=~/e/);
  LW::crawl_set_config('use_params',1)		if($G_O{c}=~/p/);
  LW::crawl_set_config('do_head',1)		if($G_O{c}=~/h/);
  LW::crawl_set_config('url_limit',10000)	if($G_O{c}=~/l/);
}
$G_VARBS{'crawl_depth'}=$G_O{d}||2;


#########################################################################

wdisplay(text=>"Whisker $G_VERSION beginning test against $G_VARBS{target}");

# so many people don't understand what do to with whisker output...
my $notice=<<EOT; 
Whisker scans for CGIs by checking to see if the server says a particular
URL exists.  However, just because a URL exists does not necessarily mean
it is vulnerable/exploitable--the vulnerability might be limited to only a
certain version of the CGI, and the server might not be using the
vulnerable version.  There is also the case where many scripts use the
same generic CGI name (like count.cgi); in this case, the exact CGI being
used may not be the same one that contains the vulnerability.

Thus, the actual vulnerability of the CGI must be verified in order to get
a true assessment of risk.  Whisker only helps in pointing out the problem
areas.  The next step after scanning with whisker is to review each found
CGI by reviewing the reference URLs or searching for the CGI name on
SecurityFocus.com or Google.com.
EOT
wdisplay(title=>'Notice',text=>$notice) if(!defined $G_VARBS{'skip_disclaimer'});

# load and run tests
rewinddir(DIR);
while(my $file=readdir(DIR)){
	require("$G_VARBS{'WHISKER_DIR'}$file") if($file=~m#\.test$#);
}
closedir(DIR);

_finish_init();
$G_VARBS{start_time}=time();
foreach (@{$G_HOOKS{'INIT'}}){ &$_; }
_exercise();
_crawler() unless (defined $G_VARBS{'skip_spider'});
foreach (@{$G_HOOKS{'TESTS'}}){ &$_; }
foreach (@{$G_HOOKS{'ANALYSIS'}}){ &$_; }
wdisplay(text=>'Whisker scan completed in '.sec2print(time()-$G_VARBS{start_time}));
foreach (@{$G_HOOKS{'EXIT'}}){ &$_; }


exit;


#########################################################################

sub sec2print {
	my ($s,$out,$hours,$mins)=(shift,'',0,0);
	$hours = int($s/3600);
	$mins = int(($s-($hours*3600))/60);
	$out="$hours hours, " if($hours>0);
	$out.="$min minutes" if($mins>0 || $hours>0);
	$out='less than 1 minute' if($mins==0 && $hours==0);
	return $out;
}

sub _init {
	LW::http_init_request(\%G_REQ);
	if(!defined $G_VARBS{'dir_codes'}){
		push(@{$G_VARBS{'dir_codes'}}, 200, 403); }
	if(!defined $G_ARRAY{'perl_exts'}){
		push(@{$G_ARRAY{'perl_exts'}}, '','.pl','.cgi'); }
	if(!defined $G_ARRAY{'php_exts'}){
		push(@{$G_ARRAY{'php_exts'}}, '.php','.php3'); }
	@{$G_ARRAY{cgibin}}=qw(cgi-bin cgi-local htbin cgi cgis cgi-win bin scripts);

	$G_REQ{'User-Agent'}="whisker/$G_VERSION";
	$G_REQ{whisker}->{'lowercase_incoming_headers'}=1;
	$G_REQ{whisker}->{'ignore_duplicate_headers'}=0;

	%{$G_VARBS{'COOKIE_JAR'}}=();
	$G_VARBS{'save_cookies'}=0;
	$G_VARBS{'DEBUG'}=0;
	$G_VARBS{'track_cookies'}=0;
	$G_VARBS{'default_test_exec'}=1;
	$G_VARBS{'rescan_lowercase'}=0;
	$G_VARBS{'options'}='h:p:S:a:P:A:I:l:T:r:t:d:c:';
	$G_VARBS{'usage'}=<<EOT;
 -h <arg>  Host to scan (www.host.com or http[s]://www.host.com:80/)
 -a <arg>  HTTP basic authorization (in the form of user:pass)	
 -P <arg>  Use HTTP proxy (in the form of proxy_host:port)
 -A <arg>  Proxy HTTP basic authorization (in the form of user:pass)	
 -t <arg>  Specify the timeout, in seconds (default: 10)
 -I <arg>  Enable the given IDS-evasive modes (see docs for mode codes)
 -S <arg>  Override the Server banner
 -G        Generic/dumb mode; force all scans in all dirs on all servers
 -r <arg>  Run only the given tests (in the form of id:id:id:...)
 -l <arg>  Log output to specified file instead of STDOUT
 -d <arg>  Depth to crawl (default: 2; main homepage counts as 1)
 -c <arg>  Crawl options:
           d  Do not crawl site
           p  Take URL parameters into account when crawling
           h  Use HEAD requests for potential speed increase
           l  Raise the URL buffer limit from 1,000 to 10,000
           a  Do not populate SPIDER directory array
           c  Add form target directories to CGIBIN directory array
           e  Do not add .asp or .php form targets to CGIBIN dir array
EOT
	$G_VARBS{'tweaks_usage'}=<<EOT;
           c  Track and handle cookies like a normal client would
           p  Include phtml in generic PHP extensions
           l  Double scan all mixed-case URLs as lowercase
           d  Suppress CGI disclaimer in output report
           i  Suppress all interactive questions
           r  Invert the -r parameter (as in, do *not* run those tests)
           R  Turn off libwhisker 'retries' feature
           0  Use HTTP/1.0 (that is a zero, and not the letter 'o')
           3  Report all move (3xx) responses
           4  Report all authentication (401) responses
           5  Report all failure (500) responses
           f  Report all form targets
           C  Report all cookies given during scanning
           A  Report everything
           D  Print debug output to STDERR
           T  Dump TRACK hash to track.log in current dir when done
EOT

}

sub _dump_track {
	my ($b,$n,$a)=('track.log','track.log',0);
	while(-e $n){
		$n=$b.$a; $a++; }
	LW::dumper_writefile($n,'track',\%G_TRACK);
}

sub _finish_init {
	my $code='sub _code_dir { my $code = shift;';
	foreach (@{$G_VARBS{'dir_codes'}}){
		$code.="return 1 if(\$code==$_);";} 
	$code.='return 0; }';
	eval $code;

	push(@{$G_HOOKS{'ANALYSIS'}},\&_record_reporter);

}

sub wgeneral {
	my ($d,$f,%o)=@_;
	return if(!defined $o{id});
	if(wtest($o{id})){
		wlog(%o) if(wexist($d,$f));
	}
}

sub wtarget {
	my $target=shift;
	LW::utils_split_uri($target,\%G_REQ);
	LW::http_fixup_request(\%G_REQ);
	@G_FOUND=();	%G_TRACK=();
}

sub warray {
	my ($name,@vals)=@_;
	return @{$G_ARRAY{$name}} if(wantarray);
	if(defined $vals[0]){
		push(@{$G_ARRAY{$name}},@vals);
	} else { delete $G_ARRAY{$name}; }
}

sub wtest {
	my $test=shift;
	my $run=$G_VARBS{'default_test_exec'};
	$run=(($run+($G_VARBS{'RUN_TESTS'}->{$test}))%2) 
		if( defined $G_VARBS{'RUN_TESTS'}->{$test});
	return $run;
}

sub wlog {
	goto &{$G_HOOKS{wlog}} if(defined $G_HOOKS{wlog});
	my %opts = @_;

	$opts{urls}=join("\n",@G_FOUND) if(!defined $opts{urls} && 
		scalar @G_FOUND >0);
	$opts{id}||='-';
	$G_IDS{$opts{id}}++;

	$opts{text}="See references for specific information on this vulnerability." 
		if(!defined $opts{text});

	$opts{reference}.="http://online.securityfocus.com/bid/$opts{bid}\n"
		if(defined $opts{bid});
	$opts{reference}.="http://cve.mitre.org/cgi-bin/cvename.cgi?name=$opts{cve}\n"
		if(defined $opts{cve});
	if(defined $opts{neo}){
		$opts{reference}.="http://archives.neohapsis.com/archives/$opts{neo}\n";
		delete $opts{neo};
	}

	wdisplay(%opts);
}

sub wdisplay {
	goto &{$G_HOOKS{wdisplay}} if(defined $G_HOOKS{wdisplay});
	my %opts = @_;

	if(defined $opts{type} && $opts{type} eq 'raw'){
		syswrite($G_VARBS{OUT},$opts{text},length($opts{text}));
		return;
	}

	delete $opts{type};
	my $out = '-'x76;
	$out.="\n";

	my @special=qw(title id bid cve severity);
	foreach (@special) { 
		if(defined $opts{$_}){
			$out.=ucfirst($_).': '.$opts{$_}."\n";
			delete $opts{$_};}}
	if(defined $opts{urls}){
		if($opts{urls}=~s/\n/\n\t/g){
			$out.="Found URLs:\n\t$opts{urls}\n";
		} else {
			$out.="Found URL: $opts{urls}\n";
		}
		delete $opts{urls};
	}

	if(defined $G_VARBS{'auto_notice'}){
		$out.="Notice: $G_VARBS{'auto_notice'}\n"; 
		delete $G_VARBS{'auto_notice'};}

	while( ($k,$v)=each %opts){
		next if($k eq 'text' || $k eq 'reference');
		$out.=ucfirst($k).": $v\n"; }

	$out.="\n";
	$out.=$opts{text}."\n" if(defined $opts{text});
	$out.="\nReferences:\n".$opts{reference}."\n"
		if(defined $opts{reference});
	$out.="\n";
	syswrite($G_VARBS{OUT},$out,length($out));
}


sub wexist {
	goto &{$G_HOOKS{wexist}} if(defined $G_HOOKS{wexist});
	my ($dirstr,$filestr)=@_;

	delete $G_VARBS{'auto_notice'};
	my (@found,@dirs);
	$dirstr=~tr/ \t//d;
	if(index($dirstr,',') >=0) {
		@dirs=split(',',$dirstr);
	} else {
		push(@dirs, $dirstr); }
	foreach (@dirs) {	
	    if($_ ne '/'){
		s;^/;;;s;/$;;; # I love perl code like this ;)
		my @dp=split('/',$_);
		LW::utils_recperm('/',0, \@dp, \@found, 
			\&wdiscoverdir, \%G_TRACK, \%G_ARRAY,
			\&_code_dir);
	    } else { push(@found,'/'); 
	}   } 
	return 0 if((scalar @found)==0); # no valid dirs

	my %req;
	_ch(\%G_REQ,\%req);
	my $F=0;
	@G_FOUND=();	
	my @EXTS=('');
	if($filestr=~s#\[(.+)\]$##){
		my $t=$1; $t=~tr/ //d;
		if(substr($t,0,1) eq '@'){
			$t=~tr/@//d;
			@EXTS=@{$G_ARRAY{lc($t).'_exts'}};
		} else {
			@EXTS=split(/,/,$t);
	}	}

	my $D; do{ $D=1;
	map {	my $d=$_; foreach my $e (@EXTS) { $e=~tr/ //d;
		my $x="$d$filestr$e";
		$x=~s#/{2,}#/#g;
		$req{whisker}->{uri}=$x;
		if(!_do_request(\%req,\%G_RESP)){
			_d_response(\%G_RESP);
			$G_TRACK{"$d$filestr$e"}=$G_RESP{whisker}->{code};
			if(_code_page($G_RESP{whisker}->{code})){
				$F++;
				push(@G_FOUND,$x); }}
	  }} @found;
	  if($filestr=~tr/A-Z// && $G_VARBS{'rescan_lowercase'}){
		$D=0; $filestr=lc($filestr); }
	} while($D==0);

	return $F;
}


sub wdiscoverdir {
	return 1 if(defined $G_VARBS{alldirs});
	my ($dir,%resp)=(shift);
	$dir=~s#/+#/#g;
	$dir.='/' if(substr($dir,-1,1) ne '/');
	return _code_dir($G_TRACK{$dir}) if(defined $G_TRACK{$dir}&&
		$G_TRACK{$dir} ne '?');
	my %req;
	_ch(\%G_REQ,\%req);
	$req{whisker}->{uri}=$dir;
	return 0 if _do_request(\%req,\%resp);
	_d_response(\%resp);
	$G_TRACK{$dir}=$resp{whisker}->{code};
	return _code_dir($resp{whisker}->{code});
}


sub wserver {
	return 1 if(defined $G_VARBS{allservers});
	my $param=shift;
	return 1 if($G_VARBS{server_banner}=~m/$param/);
	return 0;
}

sub warrayhas {
	my ($val,$name)=@_;
	foreach (@{$G_ARRAY{$name}}){
		return 1 if($val eq $_); }
	return 0;
}

sub _code_page {
	goto &{$G_HOOKS{code_page}} if(defined $G_HOOKS{code_page});
	my $code = shift;
	return 1 if($code==200);
	return 0;
}

sub _uniq_array {
	my ($name,%x)=(shift);
	return if(!defined $G_ARRAY{$name});
	foreach (@{$G_ARRAY{$name}}){ $x{$_}++; }
	@{$G_ARRAY{$name}} = keys %x;
}

sub _d_response {
	my $hr=$_[0];
	my $wr=$$hr{whisker};

	if(!defined $$wr{lowercase_incoming_headers} ||
			$$wr{lowercase_incoming_headers}==0){
		LW::utils_lowercase_headers($hr);
	}

	if(defined $$hr{'set-cookie'}){
		if($G_VARBS{'save_cookies'} || defined $G_HOOKS{d_cookie}){
			my @cookies;
			if(ref($$hr{'set-cookie'})){
				push @cookies, @{$$hr{'set-cookie'}};
			} else { push @cookies, $$hr{'set-cookie'}; }
			if(defined $G_HOOKS{d_cookie}){
				foreach $cookie (@cookies){
					&{$G_HOOKS{'d_cookie'}}($cookie); }}
			push(@{$G_RECORD{COOKIES}},@cookies) if($G_VARBS{'save_cookies'});
		}
	}

	my $C=$$wr{http_resp};
	# handle custom-404's
	if(defined $G_VARBS{'404_alternate'}){
		if($G_VARBS{'404_alternate'}==200){
			if($C==200 && LW::md5($$wr{data}) eq
					$G_VARBS{'200_as_404_data'}){
				$C=$$wr{code}=$$wr{http_resp}=404;
				print STDERR "404 $$wr{uri} (custom 404 remap)\n"
					if($G_VARBS{'DEBUG'}>0);
			}
		} else { # 302
			if($C==302 && $$hr{location} eq $G_VARBS{'302_as_404_data'}){
				$C=$$wr{code}=$$wr{http_resp}=404;
				print STDERR "404 $$wr{uri} (custom 404 remap)\n"
					if($G_VARBS{'DEBUG'}>0);
	}	}	}

	# record interesting info
	if($C==302 || $C==301 || $C==307 || $C==303){
		push(@{$G_RECORD{MOVES}},"$$wr{uri} -> $$hr{location}"); }
	if($C==401){
		push(@{$G_RECORD{AUTH}},"$$wr{uri}"); }
	if($C==500){
		push(@{$G_RECORD{FAIL}},"$$wr{uri}"); }

	if(defined $$hr{'content-length'} && $$hr{'content-length'}==0){
		$G_VARBS{'auto_notice'}='content length is 0; this could be a false/fake CGI';
	}

	if(defined $$hr{server}){
		if(!defined $G_VARBS{server_banner}){
			$G_VARBS{server_banner}=$$hr{server};
	                my $text=<<EOT;
Informational: the server returned the following banner:
        $$hr{server}
EOT
	                wlog(id=>100,text=>$text);
		} elsif($G_VARBS{server_banner} ne $$hr{server} &&
				!defined $G_IDS{107}){
	                my $text=<<EOT;
Notice!  The server banner changed during scanning to the following:
        $$hr{server}
EOT
	                wlog(id=>107,title=>'Server banner changed',text=>$text);
	}	}

	goto &{$G_HOOKS{d_response}} if(defined $G_HOOKS{d_response});
}

sub _record_reporter {
	my %unique;

	my @W=(	'COOKIES:save_cookies:The following cookies were encountered while scanning:Encountered cookies',
		'MOVES:save_moves:The following redirections were encountered while scanning:Encountered moves/redirections',
		'AUTH:report_auth:The following URLs requested authentication:Authentication requests',
		'FAIL:report_fail:The following URLs caused server failures:Encountered failures',
		'FORMTARGETS:report_ftargets:The following URLs were the targets of forms:Encountered form targets'
	);

	foreach (@W){
		my @p=split(/:/,$_);
		if(scalar @{$G_RECORD{$p[0]}}>0 && defined $G_VARBS{$p[1]}){
			%unique=();
			foreach (@{$G_RECORD{$p[0]}}){ $unique{$_}++; }
			my $text="$p[2]:\n\t";
			$text.=join("\n\t",keys %unique)."\n";
			wdisplay(text=>$text,title=>$p[3]);
		}
	}

}

sub _crawler {

	wdisplay(text=>'Whisker is currently crawling the website; please be patient.');

	LW::crawl_set_config('source_callback'=>\&_crawl_scallback,
		'save_skipped'=>1,
		'reuse_cookies'=>0);

	LW::crawl_set_config('save_cookies'=>1) if(defined $G_VARBS{'save_cookies'});
	LW::crawl_set_config('reuse_cookies'=>1) if(defined $G_VARBS{'track_cookies'});

	my $start=$G_VARBS{'target'};
	$start='http://'.$start if($start!~m#^[a-zA-Z]+://#);
	$start=~m#^([a-zA-Z]+://[^/]+)#;
	$start="$1/";

	LW::crawl($start,$G_VARBS{'crawl_depth'},\%G_TRACK,\%G_REQ);
	wdisplay(text=>'Whisker is done crawling the website.');

	my ($k,$v,%x);
	while( ($k,$v)=each %LW::crawl_forms){
		push(@{$G_RECORD{FORMTARGETS}},$k); 
		if(defined $G_VARBS{'migrate_formtargdirs'}){
			$k="$G_VARBS{'target'}/$k" if($k!~m#^[a-z]+://#i);
			my @p=LW::utils_split_uri($k);
			next if(defined $p[2] && $G_VARBS{target}!~/$p[2]/);
			next if($p[0]=~m#\.(asp|php)$# && defined $G_VARBS{'skipaspphp_formtargdirs'});
			$p[0]=~s#/[^/]*$#/#; 
			$x{$p[0]}++ if($p[0] ne '/' && $p[0] ne '');
		}
	}

	if(defined $G_VARBS{'migrate_formtargdirs'}){
		push(@{$G_ARRAY{cgibin}},keys %x);
		_uniq_array('cgibin');
	}

	my %temp=();
	@{$G_ARRAY{'SPIDER'}}=();
	while( ($k,$v)=each %G_TRACK){
		$k=~s#/[^/]*$##; $k=~s#^/##;
		next if($k eq '');
		my @p=split(/\//,$k);
		my $o='/';
		foreach (@p){
			$o.="$_/";
			$G_TRACK{$o}=200 if(!defined $G_TRACK{$o});
			$temp{$o}++ unless(defined $G_VARBS{'skip_spiderdirs'});
		}
	}

	push(@{$G_ARRAY{'spider'}}, keys %temp) 
		unless(defined $G_VARBS{'skip_spiderdirs'});

	my @nb=();
	while( ($k,$v)=each %LW::crawl_server_tags){
		next if($k eq '');
		if(!defined $G_VARBS{'server_banner'} || $G_VARBS{'server_banner'} eq ''){
			$G_VARBS{'server_banner'}=$k;
			my $text="Informational: the server returned the following banner:\n\t$G_RESP{server}\n";
			wlog(id=>100,text=>$text);
			next;
		}
		next if($k eq $G_VARBS{'server_banner'});
		push(@nb,$k);
	}

	if(scalar @nb >0){
		my $text="The server returned different banners during the course of crawling:\n\t";
		$text.= join("\n\t",@nb);
		$text.="\n";
		$text.="This could be the result of a load balancing or reverse proxy setup.\n";
		wlog(id=>103,text=>$text);
	}

	if(defined $G_VARBS{'track_cookies'} && %LW::crawl_cookies){
		push(@{$G_RECORD{'COOKIES'}}, keys %LW::crawl_cookies);
	}

}

sub _crawl_scallback {
	_d_response($_[1]);
	goto &{$G_HOOKS{d_crawl_scallback}} 
		if(defined $G_HOOKS{d_crawl_scallback});
}

sub _shutdown {
	wdisplay(text=>"Error: there was a problem connecting to the server.  The specific error returned is:\n$G_RESP{whisker}->{error}\n");
	foreach (@{$G_HOOKS{'EXIT'}}){ &$_; }
	exit;
}

sub _exercise {
	my %req;
	_ch(\%G_REQ,\%req);
	$req{whisker}->{uri}='/';
	
	&_shutdown if(_do_request(\%req,\%G_RESP));

	my $code=$G_RESP{whisker}->{code};
	if($code!=200 && $code!=302 && $code!=401){
		my $text=<<EOT;
A request to the root URL (/) on the server results in a $code response,
which is abnormal; scanning results may be skewed/affected.
EOT
		wlog(id=>104,text=>$text);
	}

	if($code==401){
		my $text=<<EOT;
The server requires authentication in order to access the root URL (/).  
It is typical in this situation that the server require authentication for
every request on the site--however whisker will still attempt to continue.
To provide authentication to the server, use the -a command line option.  
If you did provide authentication, then it's possible the username/
password you supplied were incorrect or invalid for that particular
resource.
EOT
		wlog(id=>105,text=>$text);
	}

	if(!defined $G_RESP{server}){
		my $text=<<EOT;
The server did not return a Server banner.  Whisker will attempt to
continue, but you may get better results if you use the -G
(generic) option, or force a particular Server banner via the -S option.
It's also possible that the server will be identified by future tests.
EOT
		wlog(id=>100,text=>$text);
	} else {
		my $text=<<EOT;
Informational: the server returned the following banner:
	$G_RESP{server}
EOT
		wlog(id=>100,text=>$text);
	}

	$G_VARBS{server_banner}=$G_RESP{server};

	my $random=LW::utils_randstr();
	$req{whisker}->{uri}="/$random";	
	&_shutdown if(_do_request(\%req,\%G_RESP));
	if($G_RESP{whisker}->{code}!=404){
		# Custom 404 correction code
		my $phase_two=0;
		if($G_RESP{whisker}->{code}==302){
			$phase_two=302;
			$G_VARBS{'302_as_404'}++;
			$G_VARBS{'302_as_404_data'}=$G_RESP{'location'};
		} elsif($G_RESP{whisker}->{code}==200){
			$phase_two=200;
			$G_VARBS{'200_as_404'}++;
			$G_VARBS{'200_as_404_data'}=LW::md5($G_RESP{whisker}->{data});
		} else {
			my $c=$G_RESP{whisker}->{code};
			my $text=<<EOT; 
The server returned an abnormal response ($c) to our 404 tests.  This
means whisker may not be able to determine if a CGI actually exists, and
thus the scan results are most likely going to be unreliable.
EOT
			wlog(id=>106,text=>$text);
		}
		if($phase_two){
			my (%ao,$bo);
			my %a; _ch(\%req,\%a);
			my %b; _ch(\%req,\%b);
			$a{whisker}->{uri}.='-1'.LW::utils_randstr;
			$b{whisker}->{uri}.='-2'.LW::utils_randstr;
			&_shutdown if(_do_request(\%req,\%G_RESP));
			&_shutdown if(_do_request(\%a,\%ao));
			&_shutdown if(_do_request(\%b,\%bo));

			if($phase_two==200){
				my $md0=$G_VARBS{'200_as_404_data'};
				my $md1=LW::md5($G_RESP{whisker}->{data});
				my $md2=LW::md5($ao{whisker}->{data});
				my $md3=LW::md5($bo{whisker}->{data});
				if($md1 ne $md2 || $md1 ne $md3 || $md1 ne $md0){
					my $text=<<EOT; 
The server is using a non-static custom 404 page.  This is beyond
whisker's current capabilities to cope with, and thus will make the scan
results unreliable.  This is also, technically, non-RFC compliant.
EOT
					wlog(id=>106,text=>$text);
				} else {
					# seems to be the same static 200 page
					$G_VARBS{'404_alternate'}=200;
					my $text=<<EOT; 
The server is using a static custom 404 page.  Fortunately, since it's
static, whisker should be able to accomodate it--and thus the scan results
should be unchanged for the most part.
EOT
					wlog(id=>106,text=>$text);
			}	}
			if($phase_two==302){
				my $l0=$G_VARBS{'302_as_404_data'};
				my $l1=$G_RESP{location};
				my $l2=$ao{location};
				my $l3=$bo{location};
				if($l1 ne $l2 || $l1 ne $l3 || $l1 ne $l0){
					my $text=<<EOT; 
The server issues 302 (redirect) responses instead of 404 responses, and
unfortunately, the redirect locations differ.  This is beyond whisker's
current capabilities to cope with, and thus will make the scan results
unreliable.  This is also, technically, non-RFC compliant.
EOT
					wlog(id=>106,text=>$text);
				} else {
					# seems to be the same static 200 page
					$G_VARBS{'404_alternate'}=302;
					my $text=<<EOT; 
The server seems to be issuing 302 redirects instead of 404 responses;
fortunately, all the redirects go to the same location, so whisker should
be able to accomodate it--and thus the scan results should be unchanged
for the most part.
EOT
					wlog(id=>106,text=>$text);
			}	}
	}	} # end of grody custom-404-correction code
}

sub _do_request { # wrapper for a few things
	my ($req,$resp)=@_;
	my %R;
	LW::cookie_write(\%{$G_VARBS{'COOKIE_JAR'}},$req)
		if(defined $G_VARBS{'track_cookies'});
	if(defined $G_VARBS{'ids_modes'}){
		_ch($req,\%R); $req=\%R;
		LW::anti_ids($req,$G_VARBS{'ids_modes'});
	}
	my $ret=LW::http_do_request($req,$resp);
	LW::cookie_read(\%{$G_VARBS{'COOKIE_JAR'}},$req)
		if(defined $G_VARBS{'track_cookies'});
	print STDERR "$$resp{whisker}->{code} $$req{whisker}->{uri}\n"
		if($G_VARBS{'DEBUG'}>0);
	return $ret;
}


sub _ph {
	my ($r,$o)=shift;
	print STDOUT <<EOT;
Do you need a HTTP proxy in order to reach '$$r{whisker}->{host}'?
EOT
	do {
		print "Choice [y or n]: ";
		my $x=<STDIN>;
		$x=~tr/ynYN//cd;
	} while ($x eq '');

	if($x=~/y/i){
		my $loop=1;
		do {
			print STDOUT "Enter proxy host address: ";
			my $host=<STDIN>; $host=~tr/ \t\r\n//d;
			print STDOUT "Proxy port: ";
			my $port=<STDIN>; $port=~tr/0-9//cd;
			print STDOUT "Proxy username (or leave blank if not needed): ";
			my $user=<STDIN>; $user=~tr/ \t\r\n//d;
			my $pass='';
			if($user ne ''){
			print STDOUT "Proxy password for given username: ";
				$pass=<STDIN>; $pass=~tr/\r\n//d;
			}
			print STDOUT <<EOT;
Does the following information look correct?
Proxy host:	$host
Proxy port:	$port
Proxy username:	$user
Proxy password:	$pass

Enter 'y' to continue, 'n' to go back and change something.
EOT
			do {
				print "Choice [y or n]: ";
				my $x=<STDIN>;
				$x=~tr/ynYN//cd;
			} while ($x eq '');
			$loop=0 if($x=~/y/);
		} while($loop);

		$$r{whisker}->{proxy_host}=$host;
		$$r{whisker}->{proxy_port}=$port;
		if($user ne ''){
			LW::auth_set_header('proxy-basic',$r,$user,$pass);
		}
	}
	LW::http_fixup_request($r);
	return LW::http_do_request($r,$o);
}


sub _ch { # copy whisker hash; overcomes reference problems
	my ($from,$to)=@_;
	%$to=%$from;
	my %temp=%{$$from{whisker}};
	$$to{whisker}=\%temp;
}
